import java.util.Arrays;
import java.util.Random;

import mpi.*;

/**
 * Trieda implementujuca aktivitu Master procesu.
 */
public class Master implements Runnable {

	/**
	 * Buffer na prijimanie sprav - master procesu staci pole velkosti 1.
	 */
	private int[] intBuffer = new int[1];

	/**
	 * Pocet worker procesov. Predpokladame, ze ich ranky zacinaju od 1 (0 je
	 * master).
	 */
	private int workersCount;

	/**
	 * Pole obsahujuce vstup problemu - pole vah, ktore chceme co najlepsie
	 * rozdelit
	 */
	private int[] weights;

	/**
	 * Pole, v ktorom sa generuju prefixy rozdelenia.
	 */
	private int[] partitionPrefix;

	/**
	 * Vytvori noveho master-a
	 */
	public Master() {
		workersCount = MPI.COMM_WORLD.Size() - 1;
		if (workersCount == 0) {
			throw new RuntimeException(
					"Je pozadovany aspon jeden worker proces.");
		}
	}

	@Override
	public void run() {
		// Nacitame (vygenerujeme) vstup
		loadInput();

		// Rozposleme vstup worker procesom
		sendInputToWorkers();

		// Nastavime velkost prefixu rozdeleni odosielaneho worker-om
		partitionPrefix = new int[Math.max(weights.length
				- MPJPartitionSolver.WORK_LOAD_SIZE, 0)];

		// Postupne rozposleme pracu workerom
		sendWork(0);

		// Zozbierame od workerov vysledky
		collectResults();
	}

	/**
	 * Nacita vstup - v nasom pripade si vygenerujeme nahodny.
	 */
	private void loadInput() {
		Random random = new Random();

		weights = new int[11 + random.nextInt(10)];
		for (int i = 0; i < weights.length; i++)
			weights[i] = random.nextInt(2000);
		
		logMessage("Vstup: " + Arrays.toString(weights));
	}

	/**
	 * Systemom samostanych point-to-point sprav rozposle vsetkym workerom
	 * vstup. Alternativou by bolo pouzitie metody MPI.COMM_WORLD.Bcast.
	 */
	private void sendInputToWorkers() {
		for (int i = 0; i < workersCount; i++)
			MPI.COMM_WORLD.Send(weights, 0, weights.length, MPI.INT, i + 1,
					MPJPartitionSolver.INPUT_DATA_TAG);
	}

	/**
	 * Postupne rekurzivne generuje prefixy rozdelenia od zadaneho indexu a
	 * rozposiela pracu workerom.
	 */
	private void sendWork(int fromIdx) {
		if (fromIdx == partitionPrefix.length) {
			// Cakame na ziadost o pracu od ktorehokolvek workeru
			Status status = MPI.COMM_WORLD.Recv(intBuffer, 0, intBuffer.length,
					MPI.INT, MPI.ANY_SOURCE, MPJPartitionSolver.ASK_WORK_TAG);

			// Procesu, ktory ziadal o pracu odosleme prefix rozdeleni, ktore
			// ma preskumat
			MPI.COMM_WORLD.Send(partitionPrefix, 0, partitionPrefix.length,
					MPI.INT, status.source, MPJPartitionSolver.WORK_TAG);

			return;
		}

		for (int i = 0; i < 2; i++) {
			partitionPrefix[fromIdx] = i;
			sendWork(fromIdx + 1);
		}
	}

	/**
	 * Pozbiera vysledky od worker procesov.
	 */
	private void collectResults() {
		// Caka za ziadostami o pracu od workerov. Na kazdu ziadost odpovie
		// poziadavkou na zaslanie vysledku a skoncenie.
		for (int i = 0; i < workersCount; i++) {
			Status status = MPI.COMM_WORLD.Recv(intBuffer, 0, intBuffer.length,
					MPI.INT, MPI.ANY_SOURCE, MPJPartitionSolver.ASK_WORK_TAG);
			MPI.COMM_WORLD.Send(intBuffer, 0, 0, MPI.INT, status.source,
					MPJPartitionSolver.GET_RESULT_AND_FINISH_TAG);
		}

		// Vysledky zozbierame aplikovanim kolektivnej komunikacie operaciou
		// reduce
		intBuffer[0] = Integer.MAX_VALUE;
		MPI.COMM_WORLD.Reduce(intBuffer, 0, intBuffer, 0, 1, MPI.INT, MPI.MIN,
				0);

		// Vypiseme najlepsi vysledok
		logMessage("Najlepsi vysledok:" + intBuffer[0]);
	}

	/**
	 * Vypise logovaciu spravu.
	 */
	private void logMessage(String message) {
		System.out.println("Master: " + message);
	}
}
