import mpi.*;

/**
 * Trieda implementujuca aktivitu Worker procesu.
 */
public class Worker implements Runnable {

	/**
	 * Buffer na prijimanie sprav tvorenych intami.
	 */
	private int[] intBuffer = new int[MPJPartitionSolver.MAX_INPUT_SIZE];

	/**
	 * Rank workera.
	 */
	private int rank;

	/**
	 * Vstup.
	 */
	private int[] weights;

	/**
	 * Pole, v ktorom worker generuje vsetky pridelene rozdelenia.
	 */
	private int[] partition;

	/**
	 * Rozdiel v tom najlepsom rozdeleni, aky sa podarilo najst.
	 */
	private int bestResult;

	/**
	 * Vytvori noveho workera.
	 */
	public Worker() {
		rank = MPI.COMM_WORLD.Rank();
	}

	@Override
	public void run() {
		// Prijme vstup od mastra.
		receiveInput();

		// Inicializujeme premenne, v ktorych si budeme pamatat najlepsi
		// vysledok a generovat rozdelenia
		partition = new int[weights.length];
		bestResult = Integer.MAX_VALUE;

		// Worker v nekonecnej slucke ziada pracu
		while (true) {
			// Odosle master procesu poziadavku na novu pracu
			MPI.COMM_WORLD.Send(intBuffer, 0, 0, MPI.INT, 0,
					MPJPartitionSolver.ASK_WORK_TAG);

			// Prijme od master procesu pracu
			Status status = MPI.COMM_WORLD.Recv(intBuffer, 0, intBuffer.length,
					MPI.INT, 0, MPI.ANY_TAG);

			// Ak sa prijala poziadavka na odoslanie vysledku a skoncenia, tak
			// to spravime a prerusime cyklus
			if (status.tag == MPJPartitionSolver.GET_RESULT_AND_FINISH_TAG) {
				sendResult();
				break;
			}

			// Dosla poziadavka na pracu, precitame si prefix rozdeleni, ktore
			// ma preskumat tento worker
			int prefixSize = status.Get_count(MPI.INT);
			System.arraycopy(intBuffer, 0, partition, 0, prefixSize);

			logMessage("Prijata praca: " + arrayToString(partition, prefixSize));

			// Spustime vypocet nad pridelenym prefixom rozdeleni
			analyzePartitions(prefixSize);
		}
	}

	/**
	 * Prijme od master procesu spravu so vstupom.
	 */
	private void receiveInput() {
		Status status = MPI.COMM_WORLD.Recv(intBuffer, 0, intBuffer.length,
				MPI.INT, MPI.ANY_SOURCE, MPJPartitionSolver.INPUT_DATA_TAG);

		int weightsSize = status.Get_count(MPI.INT);
		weights = new int[weightsSize];
		System.arraycopy(intBuffer, 0, weights, 0, weights.length);

		logMessage("Vstup prijaty.");
	}

	/**
	 * Odosle finalny vysledok (vyuzitim kolektivnej komunikacnej operacie
	 * reduce).
	 */
	private void sendResult() {
		intBuffer[0] = bestResult;
		MPI.COMM_WORLD.Reduce(intBuffer, 0, intBuffer, 0, 1, MPI.INT, MPI.MIN,
				0);

		logMessage("Vysledok " + bestResult + " bol odoslany do redukcie.");
	}

	/**
	 * Generuje vsetky mozne rozdelenia pocnuc zadanym indexom.
	 */
	private void analyzePartitions(int fromIdx) {
		if (fromIdx == partition.length) {
			int diff = 0;
			for (int i = 0; i < partition.length; i++)
				if (partition[i] == 0)
					diff += weights[i];
				else
					diff -= weights[i];

			bestResult = Math.min(bestResult, Math.abs(diff));
			return;
		}

		for (int i = 0; i < 2; i++) {
			partition[fromIdx] = i;
			analyzePartitions(fromIdx + 1);
		}
	}

	/**
	 * Vypise logovaciu spravu.
	 */
	private void logMessage(String message) {
		System.out.println("Worker <" + rank + ">: " + message);
	}

	/**
	 * Vrati obsah prefixu pola ako retazec.
	 */
	private static String arrayToString(int[] array, int length) {
		length = Math.min(length, array.length);

		StringBuilder sb = new StringBuilder();
		sb.append("[");
		for (int i = 0; i < length - 1; i++) {
			sb.append(array[i]);
			sb.append(", ");
		}

		if (length > 0)
			sb.append(array[length - 1]);

		sb.append("]");

		return sb.toString();
	}
}
